#include <Arduino.h>

// Define the number of valves
#define NUM_VALVES 4

// Define the pins for the valves
const int valvePins[NUM_VALVES] = {A0, A1, A2, A3};

// Define the structure for a valve
struct Valve {
  int pin;
  bool state;
  unsigned long openTime;
  unsigned long closeTime;
  unsigned long openCount;
  unsigned long closeCount;
  bool runStatus;
};

// Create an array of valves
Valve valves[NUM_VALVES];

// Function to initialize a valve
void initValve(Valve &valve, int pin) {
  valve.pin = pin;
  valve.state = false;
  valve.openTime = 0;
  valve.closeTime = 0;
  valve.openCount = 0;
  valve.closeCount = 0;
  valve.runStatus = false;
  pinMode(pin, OUTPUT);
  digitalWrite(pin, LOW);
}

// Function to turn on a valve
void valveOn(Valve &valve) {
  digitalWrite(valve.pin, HIGH);
  valve.state = true;
}

// Function to turn off a valve
void valveOff(Valve &valve) {
  digitalWrite(valve.pin, LOW);
  valve.state = false;
}

// Function to process serial input
void processInput() {
  if (Serial.available() > 0) {
    String input = Serial.readStringUntil('\n');
    input.trim();

    if (input == "startall") {
      for (int i = 0; i < NUM_VALVES; i++) {
        valves[i].runStatus = true;
      }
    } else if (input == "stopall") {
      for (int i = 0; i < NUM_VALVES; i++) {
        valves[i].runStatus = false;
        valveOff(valves[i]);
      }
    } else if (input.startsWith("start")) {
      int valveNum = input.substring(5).toInt() - 1;
      if (valveNum >= 0 && valveNum < NUM_VALVES) {
        valves[valveNum].runStatus = true;
      }
    } else if (input.startsWith("stop")) {
      int valveNum = input.substring(4).toInt() - 1;
      if (valveNum >= 0 && valveNum < NUM_VALVES) {
        valves[valveNum].runStatus = false;
        valveOff(valves[valveNum]);
      }
    } else {
      int valveNum, openValue, closeValue;
      int numParsed = sscanf(input.c_str(), "%d,%d,%d", &valveNum, &openValue, &closeValue);
      if (numParsed == 3 && valveNum >= 1 && valveNum <= NUM_VALVES && openValue > 0 && closeValue > 0) {
        valves[valveNum - 1].openTime = openValue;
        valves[valveNum - 1].closeTime = closeValue;
        valves[valveNum - 1].openCount = openValue;
        valves[valveNum - 1].closeCount = closeValue;
      }
    }
  }
}

// Function to update valve states
void updateValves() {
  for (int i = 0; i < NUM_VALVES; i++) {
    Valve &valve = valves[i];
    if (valve.runStatus) {
      if (valve.openCount > 0) {
        // Valve currently open
        valveOn(valve);
        valve.openCount--;
        if (valve.openCount == 0) {
          valveOff(valve);
          valve.closeCount = valve.closeTime;
        }
      } else if (valve.closeCount > 0) {
        // Valve currently closed
        valveOff(valve);
        valve.closeCount--;
        if (valve.closeCount == 0) {
          valveOn(valve);
          valve.openCount = valve.openTime;
        }
      } else {
        // Both closeCount and openCount == 0 ... reset to running state
        valveOn(valve);
        valve.openCount = valve.openTime;
      }
    }
  }
}

// Function to print the status of all valves
void printStatus() {
  for (int i = 0; i < NUM_VALVES; i++) {
    Valve &valve = valves[i];
    // Serial.print("Valve ");
    // Serial.print(i + 1);
    // Serial.print(": ");
    Serial.print(valve.openCount);
    Serial.print(":");
    Serial.print(valve.openTime);
    Serial.print(":");
    Serial.print(valve.closeCount);
    Serial.print(":");
    Serial.print(valve.closeTime);
    Serial.print(":");
    Serial.print(valve.runStatus ? "running" : "stopped");
    Serial.print(":");
    Serial.print(valve.state ? "open" : "closed");
    Serial.print(";");
  }
  Serial.println();
}

void setup() {
  Serial.begin(9600);
  for (int i = 0; i < NUM_VALVES; i++) {
    initValve(valves[i], valvePins[i]);
  }
}

void loop() {
  processInput();
  updateValves();
  printStatus();
  delay(1000); // 1 second delay
}