# Corby Schmitz (cschmitz@anl.gov)
# Standard board import
import board
# For digital IO pins to control motor and listen for button press
from digitalio import DigitalInOut, Direction
# For status LED
# import neopixel
# For initial BLE sleep
import time
# For Serial input
import supervisor

# Debug setting
# True = print statements enabled for all stages
# False = plottable values only
debug = False

class VALVE:
    def __init__(self, pin):
        self.valve = DigitalInOut(pin)           # Digital output for this instance
        self.valve.direction = Direction.OUTPUT  # Set pin as output
        self.valve.value = False                 # Set the valve to closed
        self.openCount = 0                       # Counter for open cycle = starts at 0
        self.closeCount = 0                      # Counter for close cycle = starts at 0
        self.openValue = 0                       # Cycle value as start for counter
        self.closeValue = 0                      # Cycle value as start for counter
        self.runStatus = False                   # Valve state = start inactive

    # Function to turn on a valve
    def on(self):
        self.valve.value = True

    # Function to turn off a valve
    def off(self):
        self.valve.value = False

    # Function to set a specific status (on or off)
    def setValveState(self, stateVal):
        self.valve.value = stateVal

    # Check if a valve is on
    def isOn(self):
        return self.valve.value

    # Reset the counters and close the valve (new directives)
    def resetCount(self):
        self.valve.value = False
        self.openCount = 0
        self.closeCount = 0

    # Set the open time of the valve
    def setOpenTime(self, seconds):
        if seconds > 1:
            self.openValue = seconds
            return True
        else:
            return False

    # Set the close time of the valve
    def setCloseTime(self, seconds):
        if seconds > 1:
            self.closeValue = seconds
            return True
        else:
            return False

    # Get the current open time of the valve (directive)
    def getOpenTime(self):
        return self.openValue

    # Get the current close time of the valve (directive)
    def getCloseTime(self):
        return int(self.closeValue)

    # Get the current counter value for open time
    def getOpenCount(self):
        return int(self.openCount)

    # Get the current counter value for close time
    def getCloseCount(self):
        return int(self.closeCount)

    # Reduce the open count by 1
    def decrementOpenCount(self):
        self.openCount -= 1

    # Reduce the close count by 1
    def decrementCloseCount(self):
        self.closeCount -= 1

    # Set the open count to the directive open value
    def resetOpenCount(self):
        self.openCount = self.openValue

    # Set the close count to the directive close value
    def resetCloseCount(self):
        self.closeCount = self.closeValue

    # Is the valve enabled to run
    def isRun(self):
        return self.runStatus

    # Enable the valve to run
    def startRun(self):
        self.runStatus = True

    # Disable valve run state, reset counts to 0 and ensure valve is off
    def stopRun(self):
        self.runStatus = False
        self.openCount = 0
        self.closeCount = 0
        self.off()

    # Print the status information for the valve
    def printValve(self):
        if self.runStatus:
            if self.valve.value:
                print(self.openCount, self.openValue, self.closeCount,
                      self.closeValue, "running", "open", sep=':', end='')
            else:
                print(self.openCount, self.openValue, self.closeCount,
                      self.closeValue, "running", "closed", sep=':', end='')
        else:
            if self.valve.value:
                print(self.openCount, self.openValue, self.closeCount,
                      self.closeValue, "stopped", "open", sep=':', end='')
            else:
                print(self.openCount, self.openValue, self.closeCount,
                      self.closeValue, "stopped", "closed", sep=':', end='')

# Set up vales with digital pins to control the state
valves = [VALVE(board.A0), VALVE(board.A1), VALVE(board.A2), VALVE(board.A3)]

# Prints the status of all valves on serial for applications to read
def printStatus():
    global valves
    for valve in valves:
        valve.printValve()
        print(";", end='')
    print()

# Function to parse commands and take appropriate action
def processInput():
    global valves
    if supervisor.runtime.serial_bytes_available:
        value = input().strip()
        if value == "":
            return
        elif value == "startall":
            for valve in valves:
                valve.startRun()
        elif value == "stopall":
            for valve in valves:
                valve.stopRun()
        elif value == "start1":
            valves[0].startRun()
        elif value == "start2":
            valves[1].startRun()
        elif value == "start3":
            valves[2].startRun()
        elif value == "start4":
            valves[3].startRun()
        elif value == "stop1":
            valves[0].stopRun()
        elif value == "stop2":
            valves[1].stopRun()
        elif value == "stop3":
            valves[2].stopRun()
        elif value == "stop4":
            valves[3].stopRun()
        elif len(value) >= 5:
            # Ensure that input is at least 5 characters #,#,# for processing
            # Validate the input for valve number 1-4 (inclusive)
            # Validate open and close values as integers with a positive value
            valveNum = 0
            openValue = 0
            closeValue = 0
            try:
                (valveID, openValueS, closeValueS) = value.split(",")
                valveNum = int(valveID)
                openValue = int(openValueS)
                closeValue = int(closeValueS)
                if not (valveNum >= 1 and valveNum <= 4):
                    raise ValueError
                if not openValue > 0:
                    raise ValueError
                if not closeValue > 0:
                    raise ValueError
            except ValueError:
                return
            valves[valveNum-1].setOpenTime(openValue)
            valves[valveNum-1].setCloseTime(closeValue)
            valves[valveNum-1].resetCount()
            if debug:
                print("Set valve: ", valveNum, "open:", openValue, "close:", closeValue)

if debug:
    print("Initiating program")
# state = False

# Core function on 1sec sleep cycle
while True:
    processInput()                              # Check for input
    for valve in valves:                        # Itterate through all valves
        if valve.isRun():                       # If valve is enabled
            if valve.getOpenCount() > 0:        # If open and counting down
                valve.decrementOpenCount()      # Decrement open count
                if valve.getOpenCount() == 0:   # If we have reached 0
                    valve.off()                 # Turn valve off
                    valve.resetCloseCount()     # Reset close counter to directive value
            if valve.getCloseCount() > 0:       # If closed and counting down
                valve.decrementCloseCount()     # Decrement close count
                if valve.getCloseCount() == 0:  # If we have reached 0
                    valve.on()                  # Turn valve on
                    valve.resetOpenCount()      # Reset open counter to directive value
            if (valve.getCloseCount() == 0 and valve.getOpenCount() == 0):  # If both open and close are 0, we need to start with open time
                valve.on()                      # Turn valve on
                valve.resetOpenCount()          # Reset open counter to directive value
    printStatus()                               # Print out status information
    time.sleep(1)                               # Sleep for 1 second and return to the top
